# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
import contextlib
import dataclasses
import inspect
import itertools
import json
import logging
import typing as t
import os
import re
from collections.abc import Mapping
from decimal import Decimal
from pathlib import Path
from socket import gethostname
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Sequence
from typing import Union

from fresco.exceptions import OptionsLoadedException

__all__ = ["Options"]

logger = logging.getLogger(__name__)
known_suffixes = {"py", "sh", "rc", "txt", "cfg", "ini", "json", "toml", "conf"}
priority_pattern = re.compile(r"([0-9]+)(?:-(.*))?$")


class Options(dict):
    """\
    Options dictionary. An instance of this is attached to each
    :class:`fresco.core.FrescoApp` instance, as a central store for
    configuration options.
    """

    _loaded_callbacks: List[Callable[["Options"], None]]
    _is_loaded = False

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.__dict__["_loaded_callbacks"] = []

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, key, value):
        self[key] = value

    def onload(self, fn: Callable) -> Callable:
        """
        Register a function to be called once ``load`` has finished populating
        the options object.
        """
        if self._is_loaded:
            fn(self)
        else:
            self._loaded_callbacks.append(fn)
        return fn

    def do_loaded_callbacks(self):
        for func in self._loaded_callbacks:
            func(self)

    def trigger_onload(self):
        """
        Mark the options object as having loaded and call any registered
        onload callbacks
        """
        if self._is_loaded:
            raise OptionsLoadedException("Options have already been loaded")
        self.do_loaded_callbacks()
        self.__dict__["_is_loaded"] = True

    def copy(self):
        return self.__class__(super().copy())

    def load(
        self,
        sources: t.Union[str, t.Iterable[t.Union[Path, str]]],
        tags: Sequence[str] = [],
        use_environ=False,
        strict=True,
        dir=None,
        trigger_onload=True,
    ) -> "Options":
        """
        Find all files matching glob pattern ``sources`` and populates the
        options object from those with matching filenames containing ``tags``.

        :param sources:
            one or more glob patterns separated by ";",
            or a list of glob patterns

        :param tags:
            a list of tags to look for in file names.

        :param use_environ: if true, environment variables matching previously
                            loaded keys will be loaded into the options object.
                            This happens after all files have been processed.
        :param strict: if true, the first file loaded is assumed to contain
                       all available option keys. Any new key found in a later
                       file will raise an error.

        Files may be in python (``.py``), json (``.json``), TOML (``.toml``)
        format. Any other files will be interpreted as simple lists of
        ```key=value`` pairs.

        Filename format
        ---------------

        The general format of filenames is:

        .. code-block:: shell

            <base>(.<priority number>-)(<tags>)(<suffix>)

        Example filenames:

        .. code-block:: shell

            # Just <base>
            .env

            # <base>.<suffix>
            .settings.toml

            # <base>.<tags>.<suffix>
            .env.dev.local..py

            # <base>.<priority>-<tags>.<suffix>
            .env.100-dev.py

        Priority number, if specified, is used to determine loading order.
        Lower numbers are loaded first. Priority numbers must be positive
        integers.

        Tags are delimited with periods,
        for example ``.env.production.py``.
        The filename ``setttings.dev.local.ini`` would be
        considered to have the tags ``('dev', 'local')``

        Where filename contain multiple tags, all tags must match for the file
        to be loaded.

        Tag names may contain the names of environment variable surrounded by
        braces, for example ``{USER}``. These will be substituted for the
        environment variable's value, with any dots or path separators replaced
        by underscores.

        The special variable ``{hostname}`` will be substituted for the current
        host's name with dots replaced by underscores.

        Files with the suffix ".sample" are unconditionally excluded.

        Loading order
        -------------

        Files are loaded in the following order:

        1. On priority number, from low to high. If the priority number is not
           given, a priority of zero is assumed

        2. Then in tag order, based on the ordering given in the
           ``tags`` parameter

           For example, calling ``options.load(..., tags=["dev", "local"])`` would
           cause a file named "settings.dev" to be loaded before one named
           "settings.local".

        3. Finally in lexicographical order.

        Environment variables, if requested, are loaded last.

        Example::

            opts = Options()
            opts.load(".env*", ["dev", "host-{hostname}", "local"])

        Would load options from files named ``.env``, ``.env.json``, ``.env.dev.py``
        and ``.env.local.py``, in that order.

        """
        if self._is_loaded:
            raise OptionsLoadedException("Options have already been loaded")

        candidates: List[Path] = []
        if dir is None:
            dir = Path(".")
        else:
            dir = Path(dir)

        if isinstance(sources, str):
            sources = [s.strip() for s in sources.split(";")]
        for source in sources:
            sourcepath = dir / Path(source)
            candidates.extend(
                p
                for p in sourcepath.parent.glob(sourcepath.name)
                if p.suffix.lower() != ".sample"
            )

        tag_substitutions = make_tag_substitutions()
        subbed_tags: list[str] = []
        for tag in tags:
            try:
                subbed_tags.append(tag.format(**tag_substitutions))
            except KeyError:
                pass
        tags = subbed_tags
        tagged_sources = []
        for path in candidates:
            priority = 0
            filename = path.name
            path_tags = []
            path_tags = [t for t in str(filename).split(".") if t][1:]
            if len(path_tags) > 0:
                # Ignore the final tag if it matches a common config file
                # extension
                if path_tags[-1].lower() in known_suffixes:
                    path_tags.pop()

            if path_tags:
                if m := priority_pattern.match(path_tags[0]):
                    priority = int(m.group(1), 10)
                    if m.group(2):
                        path_tags[0] = m.group(2)
                    else:
                        path_tags = path_tags[1:]

            if all(t in tags for t in path_tags):
                tagged_sources.append(TaggedSource(priority, path_tags, path))
            else:
                excluded = [t for t in path_tags if t not in tags]
                logger.debug(
                    f"Ignoring {path} as one or more tag does not match: {excluded=}"
                )

        matched = [
            ts.path for ts in sorted(tagged_sources, key=tagged_source_sort_key(tags))
        ]

        for path in matched:
            existing_keys = set(self.keys())
            logger.info(f"Loading config from {path}")
            if path.suffix == ".py":
                self.update_from_file(str(path))
            elif path.suffix == ".toml":
                import toml  # type: ignore

                with path.open("r") as f:
                    self.update(toml.load(f))
            elif path.suffix == ".json":
                with path.open("r") as f:
                    self.update(json.load(f))
            else:
                fullpath = path.resolve()
                with path.open("r") as f:
                    interpolations = dict(os.environ)
                    interpolations.update(self)
                    interpolations["__FILE__"] = str(fullpath)
                    interpolations["__DIR__"] = str(fullpath.parent)
                    self.update(
                        parse_key_value_pairs(
                            interpolations,
                            f,
                        )
                    )

            if existing_keys and set(self.keys()) != existing_keys:
                error_msg = (
                    f"settings file {path} created undefined options: "
                    f"{set(self.keys()) - existing_keys}"
                )
                if strict:
                    raise AssertionError(error_msg)
                else:
                    logger.warning(error_msg)

            if use_environ:
                self |= {
                    k: parse_value(self, os.environ[k]) for k in self if k in os.environ
                }

        if trigger_onload:
            self.do_loaded_callbacks()
            self.__dict__["_is_loaded"] = True
        return self

    def update_from_file(self, path, load_all=False):
        """
        Update the instance with any symbols found in the python source file at
        `path`.

        :param path: The path to a python source file
        :param load_all: If true private symbols will also be loaded into the
                         options object.
        """
        ns: Dict[str, Any] = {"__file__": path, "options": self}
        with open(path) as f:
            exec(f.read(), ns)
        if ns.get("options") is self:
            del ns["options"]
        self.update_from_dict(ns, load_all)

    def update_from_dict(self, d, load_all=False):
        """
        Update from the given list of key-value pairs.

        If ``load_all`` is True, all key-value pairs will be loaded.

        Otherwise, if the special key '__all__' is present, only those keys
        listed in __all__ will be loaded (same semantics as `from … import *`)

        Otherwise only those NOT beginning with ``_`` will be loaded.
        """
        if load_all:
            self.update(d)
        elif "__all__" in d:
            self.update((k, d[k]) for k in d["__all__"])
        else:
            self.update(
                (k, v) for k, v in d.items() if isinstance(k, str) and k and k[0] != "_"
            )

    def update_from_object(self, ob, load_all=False):
        """
        Update the instance with any symbols found in object ``ob``.

        :param load_all:
            If true private symbols will also be loaded into the options
            object.
        """
        self.update_from_dict(dict(inspect.getmembers(ob)), load_all)


@contextlib.contextmanager
def override_options(options, other=None, **kwargs):
    """
    Context manager that updates the given Options object with new values.
    On exit, the old values will be restored.

    This function is provided to assist with writing tests. It directly
    modifies the given options object and does not prevent other threads from
    accessing the modified values.
    """
    saved: list[tuple[str, t.Any]] = []
    items: t.Iterable[tuple[str, t.Any]] = []
    if isinstance(other, Mapping):
        items = ((k, other[k]) for k in other.keys())

    if kwargs:
        items = itertools.chain(items, kwargs.items())

    NOT_PRESENT = object()

    for k, v in items:
        if k in options:
            saved.append((k, options[k]))
        else:
            saved.append((k, NOT_PRESENT))

        options[k] = v

    yield options
    for k, v in saved:
        if v is NOT_PRESENT:
            del options[k]
        else:
            options[k] = v


def parse_value(
    options: Mapping,
    v: str,
    interpolation_sub=re.compile(r"\$\{([^}]*)\}|\$(\w[\w\d]*)").sub,
    int_match=re.compile(r"^[+-]?[1-9][0-9]*$").match,
    decimal_match=re.compile(r"^[+-]?[0-9]*\.[0-9]*$").match,
) -> Union[str, int, Decimal, bool]:
    def interpolate(m):
        s = m.group(1) or m.group(2)
        return str(options.get(s, m.group(0)))

    v = interpolation_sub(interpolate, v).strip()
    if int_match(v):
        return int(v)
    if decimal_match(v):
        return Decimal(v)
    if v.lower() == "true":
        return True
    if v.lower() == "false":
        return False
    if v and v[0] in "'\"" and v[-1] == v[0]:
        return v[1:-1]
    return v


def parse_key_value_pairs(options, lines: Iterable[str]):
    lines = (line.split("#", 1)[0] for line in lines)
    pairs = (line.split("=", 1) for line in lines if "=" in line)

    options = dict(options)
    values = {}
    for k, v in pairs:
        k = k.strip()
        values[k] = options[k] = parse_value(options, v)
    return values


def list_from_str(s, separator=","):
    """
    Return the value of ``s`` split into a list of times by ``separator``.
    Spaces around items will be stripped.
    """
    if isinstance(s, str):
        if not s.strip():
            return []
        return [item.strip() for item in s.split(",")]
    raise TypeError(f"Expected string, got {s!r}")


def dict_from_options(
    prefix: str, options: t.Mapping[str, Any], recursive: bool = False
) -> t.Dict[str, Any]:
    """
    Create a dictionary containing all items in ``options`` whose keys start
    with ``prefix``, with that prefix stripped.

    Example:

        >>> from fresco.options import Options, dict_from_options
        >>> options = Options(
        ...      {"DATABASE_HOST": "127.0.0.1", "DATABASE_USER": "scott"}
        ... )
        >>> dict_from_options("DATABASE_", options)
        {'HOST': '127.0.0.1', 'USER': 'scott'}

    If ``recursive`` is True, a recursive splitting will be
    applied. The last character of ``prefix`` will be used as the separator,
    for example::

        >>> from fresco.options import dict_from_options
        >>> options = Options(
        ...      {
        ...          "DATABASE_DEV_HOST": "127.0.0.1",
        ...          "DATABASE_DEV_USER": "scott",
        ...          "DATABASE_PROD_HOST": "192.168.0.78",
        ...          "DATABASE_PROD_HOST": "tiger",
        ...      }
        ... )
        >>> dict_from_options("DATABASE_", options, recursive=True)
        {'DEV': {'HOST': '127.0.0.1', ...}, 'PROD': {'HOST', ...}}

    """

    prefixlen = len(prefix)
    d = {k[prefixlen:]: v for k, v in options.items() if k.startswith(prefix)}

    def recursive_split(d: t.Mapping[str, Any], sep: str) -> t.Dict[str, Any]:
        d_: t.Dict[str, Any] = {}
        for k in d:
            if sep in k:
                ks = k.split(sep)
                subdict = d_
                for subkey in ks[:-1]:
                    subdict = subdict.setdefault(subkey, {})
                subdict[ks[-1]] = d[k]
            else:
                d_[k] = d[k]
        return d_

    if recursive:
        return recursive_split(d, prefix[-1])
    return d


@dataclasses.dataclass
class TaggedSource:
    priority: int
    tags: list[str]
    path: Path


def tagged_source_sort_key(
    tags: list[str],
) -> Callable[[TaggedSource], tuple[int, list[int], str]]:
    """
    Return a function that can be used as the ``key`` argument of``sorted``,
    that sorts TaggedSource objects based on
    priority, tag order, then filename.
    """

    def _sort_key(ts: TaggedSource) -> tuple[int, list[int], str]:
        sorted_tags = [tags.index(t) for t in ts.tags]
        return (ts.priority, sorted_tags, str(ts.path))

    return _sort_key


def make_tag_substitutions():
    """
    Return a dict of substitutions to make in tag names, based on the current
    hostname and environment variables

    This permits tags to be specified as, for example,
    '{hostname}' or '{FRESCO_PROFILE}'
    and the associated value will be substituted in.

    Path separators and dots (which separate tags) are replaced in substitution
    values by an underscore.
    """
    substitutions = os.environ.copy()
    substitutions["hostname"] = gethostname()
    substitutions = {
        k: v.replace(".", "_").replace(os.pathsep, "_")
        for k, v in substitutions.items()
    }
    return substitutions



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from fresco.decorators import json_response


class TestJSONResponse(object):
    def test_it_allows_custom_formatting(self):
        r = json_response({"l": [1, 2, 3]}, indent=1, separators=(", ", ": "))
        assert list(r.content_iterator) == [b'{\n "l": [\n  1, \n  2, \n  3\n ]\n}']

    def test_it_acts_as_a_filter(self):
        r = json_response({"l": [1, 2, 3]})
        assert list(r.content_iterator) == [b'{"l":[1,2,3]}']
        assert r.get_header("content-type") == "application/json"

    def test_it_acts_as_a_decorator(self):
        @json_response()
        def f():
            return {"l": [1, 2, 3]}

        assert list(f().content_iterator) == [b'{"l":[1,2,3]}']

    def test_it_acts_as_a_decorator_without_arguments(self):
        @json_response
        def f():
            return {"l": [1]}

        assert list(f().content_iterator) == [b'{"l":[1]}']



# encoding: UTF-8
from fresco.core import FrescoApp
from fresco.routing import GET
from fresco.routing import Route
from fresco.requestcontext import context
from fresco.response import Response
from fresco.routeargs import GetArg, routearg
from fresco.subrequests import subrequest
from fresco.subrequests import subrequest_bytes
from fresco.subrequests import subrequest_raw
from fresco.subrequests import subrequest_str

from unittest.mock import Mock
import typing as t
import pytest


class TestSubRequest(object):
    def test_it_raises_exception_if_decoding_impossible(self):
        def view_bytes():
            return Response(b"caf\xe9", content_type="text/plain")

        with FrescoApp().requestcontext():
            with pytest.raises(ValueError):
                subrequest(view_bytes)

    def test_it_decodes_response_content(self):
        def view_latin1():
            return Response(
                "café".encode("latin1"),
                content_type="text/plain; charset=Latin-1",
            )

        def view_utf8():
            return Response(
                "café".encode("UTF-8"),
                content_type="text/plain; charset=UTF-8",
            )

        def view_string():
            return Response("café", content_type="text/plain; charset=UTF-8")

        def view_list_of_strings():
            return Response(["café"], content_type="text/plain; charset=UTF-8")

        with FrescoApp().requestcontext():
            assert subrequest(view_utf8) == "café"
            assert subrequest(view_latin1) == "café"
            assert subrequest(view_string) == "café"
            assert subrequest(view_list_of_strings) == "café"

    def test_it_returns_a_markup_string(self):
        def view_html():
            return Response("<html>", content_type="text/html; charset=UTF-8")

        def view_text():
            return Response("text", content_type="text/plain; charset=UTF-8")

        with FrescoApp().requestcontext():
            assert hasattr(subrequest(view_html), "__html__")
            assert not hasattr(subrequest(view_text), "__html__")

    def test_it_returns_raw_response(self):
        r = Response("foo")
        with FrescoApp().requestcontext():
            assert subrequest_raw(lambda: r) is r

    def test_it_returns_content_as_bytes(self):
        def view():
            return Response(
                "café".encode("latin1"),
                content_type="text/plain; charset=Latin-1",
            )

        with FrescoApp().requestcontext():
            assert subrequest_bytes(view) == b"caf\xe9"

    def test_it_calls_response_onclose(self):
        m = Mock()
        r = Response(onclose=m)
        with FrescoApp().requestcontext():
            assert subrequest(lambda: r) == ""
        assert m.call_count == 1

    def test_viewspec_resolves(self):
        def view():
            return Response("foo")

        app = FrescoApp()
        app.route("/view", GET, view, name="named view")
        with app.requestcontext():
            assert subrequest("named view") == "foo"

    def test_viewspec_can_be_a_path(self):
        def view():
            return Response("foo")

        app = FrescoApp()
        app.route("/view", GET, view, name="named view")
        with app.requestcontext():
            assert subrequest("/view") == "foo"

    def test_viewspec_uses_routeargs(self):
        def view(a, b):
            return Response(["*" + a + b + "*"])

        app = FrescoApp()
        app.route("/view", GET, view, a=GetArg(), b=routearg(lambda r: r.method))

        with app.requestcontext("/view?a=foo"):
            assert subrequest(view, _resolve=True) == "*fooGET*"
            assert subrequest(view, _resolve=True, a="bar", b="baz") == "*barbaz*"

    def test_viewspec_resolves_dynamic(self):
        class Views(object):
            def __init__(self, a):
                self.a = a

            def view(self, b):
                return Response("*" + self.a + b + "*")

            __routes__ = [Route("/<b:str>", GET, "view", name="named view")]

        app = FrescoApp()
        app.delegate("/<a:str>", Views, dynamic=True, name="named collection")

        with app.requestcontext():
            assert (
                subrequest("named collection:named view", a="foo", b="bar")
                == "*foobar*"
            )

    def test_it_does_a_full_request(self):
        request_hook = Mock(return_value=None)
        view = Mock(return_value=Response())

        app = FrescoApp()
        app.route("/view", GET, view)
        app.process_request(request_hook)
        with app.requestcontext():
            subrequest(view)
            assert view.call_count == 1
            assert request_hook.call_count == 0

            subrequest(view, _full=True)
            assert view.call_count == 2
            assert request_hook.call_count == 1

    def test_it_maintains_the_environ(self):
        subreq_environ = None

        def view():
            nonlocal subreq_environ
            subreq_environ = context.request.environ
            return Response()

        app = FrescoApp()
        app.route("/view", GET, view)

        with app.requestcontext("/xyzzy") as c:
            original_environ = c.request.environ
            subrequest(view)
            assert subreq_environ is original_environ

    def test_it_passes_original_environ_values(self):
        subreq_environ = None

        def view():
            nonlocal subreq_environ
            subreq_environ = context.request.environ
            return Response()

        app = FrescoApp()
        app.route("/view", GET, view)
        environ_copied = {
            "HTTP_COOKIE": "1234",
            "REMOTE_IP": "192.168.0.1",
            "REMOTE_USER": "username",
        }
        environ_not_copied = {
            "PATH_INFO": "/not-copied",
            "QUERY_STRING": "not-copied",
        }

        with app.requestcontext("/xyzzy") as c:
            original_environ = c.request.environ
            c.request.environ.update(environ_copied)
            c.request.environ.update(environ_not_copied)
            subrequest(view, _full=True, _query="test")
            assert isinstance(subreq_environ, dict)
            assert subreq_environ is not original_environ
            assert all(subreq_environ[k] == environ_copied[k] for k in environ_copied)
            assert all(
                subreq_environ[k] != environ_not_copied[k] for k in environ_not_copied
            )
            assert subreq_environ["PATH_INFO"] == "/view"
            assert subreq_environ["QUERY_STRING"] == "test"

    def test_it_merges_custom_environ_keys(self):
        subreq_environ: t.Optional[dict] = None

        def view():
            nonlocal subreq_environ
            subreq_environ = context.request.environ
            return Response()

        app = FrescoApp()
        app.route("/view", GET, view)

        with app.requestcontext("/xyzzy") as c:
            original_environ = c.request.environ
            subrequest(view, _full=True, _env={"hey.there": True})
            assert subreq_environ is not None
            assert "hey.there" in subreq_environ
            assert "hey.there" not in original_environ

    def test_it_exhausts_content_iterator_before_closing_response(self):
        tracker = ["view inited"]

        def view():
            def responder():
                tracker.append("responder started")
                yield ",".join(tracker)

            r = Response(responder())
            r = r.add_onclose(lambda: tracker.append("response closed"))
            return r

        app = FrescoApp()
        app.route("/", GET, view)

        with app.requestcontext("/"):
            result = subrequest_str(view, _full=True)
        assert ",".join(tracker) == "view inited,responder started,response closed"

        assert result == "view inited,responder started"



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
"""
Utilities functions for URL building and manipulation
"""

from collections.abc import Mapping
from functools import lru_cache
from itertools import chain
from typing import List
from urllib.parse import quote_plus
from urllib.parse import urlparse
from urllib.parse import urlunparse
from urllib.parse import urlencode
from urllib.parse import parse_qsl
import posixpath
import re
import typing as t

from fresco.defaults import DEFAULT_CHARSET
from fresco.multidict import MultiDict
from fresco.types import QuerySpec

__all__ = "join_path", "url_join", "strip_trailing_slashes", "normpath", "add_query"


def join_path(a, b):
    """
    Join two URL path segments together.
    Unlike ``os.path.join``, this removes leading slashes before joining the
    paths, so absolute paths are appended as if they were relative.

    Synopsis::

        >>> join_path('/', '/abc')
        '/abc'

        >>> join_path('/', '/abc/')
        '/abc/'

        >>> join_path('/foo', '/bar/')
        '/foo/bar/'

        >>> join_path('/foo/', '/bar/')
        '/foo/bar/'

        >>> join_path('/foo', 'bar')
        '/foo/bar'
    """

    if not b:
        return a

    while a and a[-1] == "/":
        a = a[:-1]

    while b and b[0] == "/":
        b = b[1:]

    return a + "/" + b


def url_join(base, rel):
    """
    Examples::

        >>> from fresco.util.urls import url_join
        >>> url_join('http://example.org/', 'http://example.com/')
        'http://example.com/'
        >>> url_join('http://example.com/', '../styles/main.css')
        'http://example.com/styles/main.css'
        >>> url_join('http://example.com/subdir/', '../styles/main.css')
        'http://example.com/styles/main.css'
        >>> url_join('http://example.com/login', '?error=failed+auth')
        'http://example.com/login?error=failed+auth'
        >>> url_join('http://example.com/login', 'register')
        'http://example.com/register'
    """
    prel = urlparse(rel)

    # Link is already absolute, return it unchanged
    if prel.scheme:
        return rel

    pbase = urlparse(base)
    path = pbase.path
    if prel.path:
        path = normpath(posixpath.join(posixpath.dirname(pbase.path), prel.path))

    return urlunparse(
        (
            pbase.scheme,
            pbase.netloc,
            path,
            prel.params,
            prel.query,
            prel.fragment,
        )
    )


def strip_trailing_slashes(path):
    """\
    Remove trailing slashes from the given path.

    :param path: a path, eg ``'/foo/bar/'``
    :return: The path with trailing slashes removed, eg ``'/foo/bar'``
    """
    while path and path[-1] == "/":
        path = path[:-1]
    return path


@lru_cache(50)
def normpath(path):
    """
    Return ``path`` normalized to remove '../' etc.

    This differs from ``posixpath.normpath`` in that:

    * trailing slashes are preserved
    * multiple consecutive slashes are always condensed to a single slash

    Examples::

        >>> normpath('/hello/../goodbye')
        '/goodbye'
        >>> normpath('//etc/passwd')
        '/etc/passwd'
        >>> normpath('../../../../etc/passwd')
        'etc/passwd'
        >>> normpath('/etc/passwd/')
        '/etc/passwd/'
    """
    segments = path.split("/")
    newpath: List[str] = []
    newpath_append = newpath.append
    last = len(segments) - 1
    for pos, seg in enumerate(segments):
        if seg == "" or seg == ".":
            seg = ""
            allow_empty = (
                pos == 0
                or pos == last
                and newpath
                and newpath[-1] != ""
                or pos == last
                and newpath == [""]
            )
            if not allow_empty:
                continue

        elif seg == "..":
            if newpath and newpath != [""]:
                newpath.pop()
            continue

        newpath_append(seg)

    return "/".join(newpath)


def make_query(
    data: QuerySpec = [],
    separator: str = "&",
    charset: t.Optional[str] = None,
    **kwargs: t.Any,
) -> str:
    """
    Return a query string formed from the given dictionary data.

    If no encoding is given, unicode values are encoded using the character set
    specified by ``fresco.defaults.DEFAULT_CHARSET``.

    Basic usage::

        >>> make_query({'search': 'foo bar', 'type': u'full'})
        'search=foo+bar&type=full'

        >>> make_query(search='foo', type=u'quick')
        'search=foo&type=quick'

    Changing the separator to ';'::

        >>> make_query({'search': 'foo', 'type': u'full'}, separator=';')
        'search=foo;type=full'

    Multiple values can be provided per key::

        >>> make_query(search=['foo', u'bar'])
        'search=foo&search=bar'

    Exclude values by setting them to ``None``::

        >>> make_query(x=1, y=None)
        'x=1'

    :param data: data for query string
    :param separator: separator used between each key value pair
    :param charset: encoding to be used for unicode values
    :rtype: str
    """
    items: t.Iterable[tuple[str, t.Any]]
    if isinstance(data, MultiDict):
        items = data.allitems()
    elif isinstance(data, Mapping):
        items = data.items()  # type: ignore
    else:
        items = data
    if kwargs:
        items = chain(items, kwargs.items())

    if charset is None:
        charset = DEFAULT_CHARSET

    pairs: list[str] = []
    append = pairs.append
    for k, v in items:
        if isinstance(v, (str, bytes)):
            append(f"{quote_plus(k, charset)}={quote_plus(v, charset)}")
        elif v is None:
            pass
        elif hasattr(v, "__iter__"):
            for v in v:
                append(f"{quote_plus(k, charset)}={quote_plus(str(v), charset)}")
        else:
            append(f"{quote_plus(k, charset)}={quote_plus(str(v), charset)}")
    return separator.join(pairs)


def _qs_frag(key, value, charset=None):
    """
    Return a fragment of a query string in the format 'key=value'::

        >>> _qs_frag('search-by', 'author, editor')
        'search-by=author%2C+editor'

    If no encoding is specified, unicode values are encoded using the character
    set specified by ``fresco.defaults.DEFAULT_CHARSET``.

    :rtype: str
    """
    if charset is None:
        charset = DEFAULT_CHARSET

    key = str(key).encode(charset)
    value = str(value).encode(charset)

    return quote_plus(key) + "=" + quote_plus(value)


_valid_url_chars = re.compile(
    r"^(?:[A-Za-z0-9\-._~:/?#\[\]@!$&'\(\)\*\+,;=']+|%[0-9A-Fa-f]{2})*$"
)


def is_safe_url(
    url,
    allowed_hosts=frozenset(),
    allowed_schemes={"http", "https"},
    valid_url=_valid_url_chars.match,
    urlparse=urlparse,
):
    """
    Return True if the given URL is for a trusted origin and can safely be
    redirected to.

    :param allowed_hosts: a list of hosts considered trusted. If not supplied
                          the current HTTP_HOST or SERVER_NAME/SERVER_PORT are
                          used.
    :param allowed_schemes: a list of permitted URL schemes. Defaults to
                            ``{'http', 'https'}``
    """
    if not url:
        return False
    if url.startswith("///"):
        return False
    if not valid_url(url):
        return False

    if not allowed_hosts:
        from fresco import context

        env = context.request.environ
        try:
            http_host = env["HTTP_HOST"]
            if ":" in http_host:
                h_host, h_port = http_host.split(":")
                h_scheme = env["wsgi.url_scheme"]
                if (h_scheme == "https" and h_port == "443") or (
                    h_scheme == "http" and h_port == "80"
                ):
                    allowed_hosts = {h_host}
                else:
                    allowed_hosts = {http_host}
            else:
                allowed_hosts = {http_host}
        except KeyError:
            server_name = env["SERVER_NAME"]
            server_port = env["SERVER_PORT"]
            server_scheme = env["wsgi.url_scheme"]
            if (server_scheme == "https" and server_port == "443") or (
                server_scheme == "http" and server_port == "80"
            ):
                allowed_hosts = {server_name}
            else:
                allowed_hosts = {f"{server_name}:{server_port}"}

    parsed = urlparse(url)
    scheme = parsed.scheme
    netloc = parsed.netloc
    if ":" in netloc:
        host, port = netloc.split(":")
        if (scheme == "http" and port == "80") or (scheme == "https" and port == "443"):
            netloc = host

    return (
        (scheme in allowed_schemes and netloc in allowed_hosts)
        or (scheme == "" and netloc == "")
        or (scheme == "" and netloc in allowed_hosts)
    )


def add_query(
    url: str,
    data: t.Union[t.Sequence[tuple[str, t.Any]], t.Mapping[str, t.Any]] = [],
    **kwargs: t.Any,
) -> str:
    """
    Return the given URL with the given query data appended

    :param url:
        The base URL
    :param data:
        Optional data items, supplied either as a mapping or list of (key,
        value) tuples
    :param kwargs:
        Query data supplied as keyword arguments
    """
    parsed = urlparse(url)
    return urlunparse(
        parsed._replace(
            query=urlencode(
                parse_qsl(parsed.query)
                + list(dict(data).items())
                + list(kwargs.items())
            )
        )
    )



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from fresco.request import Request
from fresco.request import currentrequest
from fresco.requestcontext import context
from fresco.response import Response
from fresco.core import FrescoApp
from fresco.core import urlfor
from fresco.defaults import DEFAULT_CHARSET
from fresco.options import Options
from fresco.routing import Route
from fresco.routing import RRoute
from fresco.routing import RouteCollection
from fresco.routing import DelegateRoute
from fresco.routing import routefor
from fresco.routing import ALL_METHODS
from fresco.routing import GET
from fresco.routing import HEAD
from fresco.routing import POST
from fresco.routing import PUT
from fresco.routing import DELETE
from fresco.routing import OPTIONS
from fresco.routing import TRACE
from fresco.routing import CONNECT
from fresco.routing import VERSION_CONTROL
from fresco.routing import REPORT
from fresco.routing import CHECKOUT
from fresco.routing import CHECKIN
from fresco.routing import UNCHECKOUT
from fresco.routing import MKWORKSPACE
from fresco.routing import UPDATE
from fresco.routing import LABEL
from fresco.routing import MERGE
from fresco.routing import BASELINE_CONTROL
from fresco.routing import MKACTIVITY
from fresco.routing import ORDERPATCH
from fresco.routing import ACL
from fresco.routing import SEARCH
from fresco.routing import PATCH
from fresco.routeargs import routearg
from fresco.routeargs import FormArg
from fresco.routeargs import PostArg
from fresco.routeargs import QueryArg
from fresco.routeargs import GetArg
from fresco.routeargs import CookieArg
from fresco.routeargs import SessionArg
from fresco.routeargs import RequestObject
from fresco.routeargs import FormData
from fresco.routeargs import PostData
from fresco.routeargs import QueryData
from fresco.routeargs import GetData
from fresco.middleware import XForwarded
from fresco.subrequests import subrequest
from fresco.subrequests import subrequest_bytes
from fresco.subrequests import subrequest_raw
from fresco.util.common import object_or_404


__version__ = "3.8.0"
__all__ = [
    "Request",
    "currentrequest",
    "context",
    "Response",
    "FrescoApp",
    "urlfor",
    "ALL_METHODS",
    "DEFAULT_CHARSET",
    "GET",
    "HEAD",
    "POST",
    "PUT",
    "DELETE",
    "OPTIONS",
    "TRACE",
    "CONNECT",
    "VERSION_CONTROL",
    "REPORT",
    "CHECKOUT",
    "CHECKIN",
    "UNCHECKOUT",
    "MKWORKSPACE",
    "UPDATE",
    "LABEL",
    "MERGE",
    "BASELINE_CONTROL",
    "MKACTIVITY",
    "ORDERPATCH",
    "ACL",
    "SEARCH",
    "PATCH",
    "DelegateRoute",
    "Options",
    "Route",
    "RouteCollection",
    "RRoute",
    "routearg",
    "FormArg",
    "PostArg",
    "QueryArg",
    "GetArg",
    "CookieArg",
    "SessionArg",
    "RequestObject",
    "FormData",
    "PostData",
    "QueryData",
    "GetData",
    "routefor",
    "XForwarded",
    "subrequest",
    "subrequest_bytes",
    "subrequest_raw",
    "object_or_404",
]



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from unittest.mock import Mock, call

from fresco import routeargs
from fresco.core import FrescoApp
from fresco.response import Response
from fresco.routing import GET
from fresco.routing import POST
from fresco.routing import Route


class TestRouteArg(object):
    def test_routekwarg_configured(self):
        A = routeargs.RouteArg()
        route = Route("/", GET, lambda r: None, x=A)
        assert A.route is route
        assert A.name == "x"

    def test_routekwarg_value_passed(self):
        view = Mock(return_value=Response())
        routekwarg = Mock(spec=routeargs.RouteArg)
        routekwarg.return_value = "xyzzy"

        app = FrescoApp()
        app.route("/", GET, view, x=routekwarg)
        with app.requestcontext("/") as c:
            app.view()
            assert routekwarg.call_args_list == [call(c.request)]
            assert view.call_args_list == [call(x="xyzzy")]

    def test_queryarg_value_passed(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.QueryArg())
        with app.requestcontext("/?x=foo"):
            app.view()
            assert view.call_args_list == [call(x="foo")]

    def test_formarg_value_passed(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.FormArg())
        with app.requestcontext("/?x=foo"):
            app.view()
            assert view.call_args_list == [call(x="foo")]

    def test_sessionarg_value_passed(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.SessionArg())
        with app.requestcontext("/?x=foo") as c:
            c.request.environ[c.request.SESSION_ENV_KEY] = {"x": "foo"}
            app.view()
            assert view.call_args_list == [call(x="foo")]

    def test_cookiearg_value_passed(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.CookieArg())
        with app.requestcontext("/", HTTP_COOKIE="x=foo"):
            app.view()
            assert view.call_args_list == [call(x="foo")]

    def test_cookiearg_listvalue_passed(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.CookieArg([str]))
        with app.requestcontext("/", HTTP_COOKIE="x=foo;x=bar"):
            app.view()
            assert view.call_args_list == [call(x=["foo", "bar"])]

    def test_requestarg_value_converted(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.FormArg(float))
        with app.requestcontext("/?x=0"):
            app.view()
            assert view.call_args_list == [call(x=0.0)]

    def test_it_converts_lists(self):
        arg = routeargs.RequestArg(converter=[str])
        arg.configure(None, "foo")
        with FrescoApp().requestcontext("/?foo=a&foo=b") as r:
            assert arg(r.request) == ["a", "b"]

    def test_requestarg_default_value(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.FormArg(default="d"))
        with app.requestcontext("/"):
            app.view()
            assert view.call_args_list == [call(x="d")]

    def test_it_doesnt_convert_default_values(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.FormArg(int, default=None))
        with app.requestcontext("/"):
            app.view()
            assert view.call_args_list == [call(x=None)]

    def test_access_to_missing_requestarg_returns_badrequest(self):
        def view(x):
            x == 0

        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.FormArg())
        with app.requestcontext("/"):
            assert "Bad Request" in app.view().status

    def test_missing_requestarg_with_conversion_returns_badrequest(self):
        def view(x):
            x == 0

        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.FormArg(int))
        with app.requestcontext("/"):
            assert "Bad Request" in app.view().status

    def test_routeargs_work_as_positional_arguments(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", GET, view, args=[routeargs.FormArg(key="x")])
        with app.requestcontext("/?x=foo"):
            app.view()
            assert view.call_args_list == [call("foo")]

    def test_routearg_classes_are_auto_instantiated(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.FormArg)
        with app.requestcontext("/?x=foo"):
            app.view()
            assert view.call_args_list == [call(x="foo")]


class TestRequestObject(object):
    def test_it_passes_the_request(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.RequestObject())
        with app.requestcontext("/") as c:
            app.view()
            x = view.call_args[1]["x"]
            assert x is c.request


class Test_routearg(object):
    def test_it_calls_func(object):
        view = Mock(return_value=Response())
        argfunc = Mock()
        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.routearg(argfunc))
        with app.requestcontext("/"):
            app.view()
            assert argfunc.call_count == 1, argfunc.call_count
            assert view.call_args[1]["x"] is argfunc()

    def test_it_passes_additional_args(object):
        view = Mock(return_value=Response())
        argfunc = Mock()
        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.routearg(argfunc, "x", y=42))
        with app.requestcontext("/") as c:
            app.view()
            assert argfunc.call_args == ((c.request, "x"), {"y": 42})
            assert view.call_args[1]["x"] is argfunc()


class TestFormData(object):
    def test_it_passes_the_form_dict(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", GET, view, x=routeargs.FormData())
        with app.requestcontext("/") as c:
            app.view()
            x = view.call_args[1]["x"]
            assert x is c.request.form


class TestJSONPayload(object):
    def test_it_passes_json_data(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", POST, view, x=routeargs.JSONPayload())
        with app.requestcontext_post(
            "/", data=b'{"foo":"bar"}', CONTENT_TYPE="application/json"
        ):
            app.view()
            x = view.call_args[1]["x"]
            assert x == {"foo": "bar"}

    def test_it_returns_bad_request_on_invalid_payload(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", POST, view, x=routeargs.JSONPayload())
        with app.requestcontext_post(
            "/", data=b"invalid json!", CONTENT_TYPE="application/json"
        ):
            r = app.view()
            assert r.status_code == 400

    def test_it_returns_a_default_value(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", POST, view, x=routeargs.JSONPayload(default="default value"))
        with app.requestcontext_post("/"):
            app.view()
        x = view.call_args[1]["x"]
        assert x == "default value"



# encoding=UTF-8
# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from io import BytesIO
import datetime

import pytest

from fresco import FrescoApp, exceptions
from fresco.tests import fixtures

context = FrescoApp().requestcontext


class TestRequestProperties(object):
    def test_url_script_name_only(self):
        with context(SCRIPT_NAME="/foo/bar", PATH_INFO="") as c:
            assert c.request.url == "http://localhost/foo/bar"

    def test_url_script_name_path_info(self):
        with context(SCRIPT_NAME="/foo/bar", PATH_INFO="/baz") as c:
            assert c.request.url == "http://localhost/foo/bar/baz"

    def test_url_normalizes_host_port(self):
        with context(HTTP_HOST="localhost:80") as c:
            assert c.request.url == "http://localhost/"
        with context(HTTP_HOST="localhost:81") as c:
            assert c.request.url == "http://localhost:81/"

    def test_url_normalizes_host_ssl_port(self):
        with context(
            environ={"wsgi.url_scheme": "https"}, HTTP_HOST="localhost:443"
        ) as c:
            assert c.request.url == "https://localhost/"

    def test_url_ignores_server_port_if_host_header_present(self):
        with context(
            environ={"wsgi.url_scheme": "https", "SERVER_PORT": "81"},
            HTTP_HOST="localhost",
        ) as c:
            assert c.request.url == "https://localhost/"

    def test_as_string(self):
        with context("http://example.org/foo?bar=baz") as c:
            assert str(c.request) == "<Request GET http://example.org/foo?bar=baz>"

    def test_url_returns_full_url(self):
        with context("http://example.org/foo?bar=baz") as c:
            assert c.request.url == "http://example.org/foo?bar=baz"

    def test_path_returns_correct_path_when_script_name_empty(self):
        with context(SCRIPT_NAME="", PATH_INFO="/foo/bar") as c:
            assert c.request.path == "/foo/bar"

    def test_path_returns_correct_path(self):
        with context(SCRIPT_NAME="/foo", PATH_INFO="/bar") as c:
            assert c.request.path == "/foo/bar"

    def test_query_decodes_unicode(self):
        with context("/?q=%C3%A0") as c:
            assert c.request.query["q"] == b"\xc3\xa0".decode("utf8")

    def test_form_decodes_unicode(self):
        with context("/?q=%C3%A0") as c:
            assert c.request.form["q"] == b"\xc3\xa0".decode("utf8")

    def test_body_decodes_unicode(self):
        with context(
            wsgi_input=b"\xc3\xa0",
            CONTENT_LENGTH="2",
            CONTENT_TYPE="text/plain; charset=UTF-8",
        ) as c:
            assert c.request.body == b"\xc3\xa0".decode("UTF-8")

    def test_body_raises_bad_request_for_invalid_encoding(self):
        with context(
            wsgi_input=b"\xc3a",
            CONTENT_LENGTH="2",
            CONTENT_TYPE="text/plain; charset=UTF-8",
        ) as c:
            with pytest.raises(exceptions.BadRequest):
                c.request.body

    def test_body_bytes_does_not_decode(self):
        with context(
            wsgi_input=b"\xc3a",
            CONTENT_LENGTH="2",
            CONTENT_TYPE="text/plain; charset=UTF-8",
        ) as c:
            assert c.request.body_bytes == b"\xc3a"

    def test_get_json_decodes_json(self):
        with context(
            wsgi_input=b'{"foo": "bar"}',
            CONTENT_LENGTH="14",
            CONTENT_TYPE="application/json",
        ) as c:
            assert c.request.get_json() == {"foo": "bar"}

    def test_get_json_ignores_mime_type(self):
        with context(
            wsgi_input=b'{"foo": "bar"}',
            CONTENT_LENGTH="14",
            CONTENT_TYPE="application/broken",
        ) as c:
            assert c.request.get_json() == {"foo": "bar"}

    def test_get_json_passes_args_to_decoder(self):
        with context(
            wsgi_input=b'{"foo": 1}',
            CONTENT_LENGTH="10",
            CONTENT_TYPE="application/broken",
        ) as c:
            assert c.request.get_json(parse_int=lambda s: s + "!") == {"foo": "1!"}

    def test_get_required_raises_badrequest(self):
        with context(
            QUERY_STRING="x=10",
            CONTENT_LENGTH="10",
            CONTENT_TYPE="application/broken",
        ) as c:
            req = c.request
            assert req.get("x", required=True) == "10"
            with pytest.raises(exceptions.BadRequest):
                req.get("y", required=True)

    def test_get_does_type_conversion(self):
        with context(
            QUERY_STRING="x=10",
            CONTENT_LENGTH="10",
            CONTENT_TYPE="application/broken",
        ) as c:
            req = c.request
            assert req.get("x") == "10"
            assert req.get("x", type=int) == 10
            assert req.get("y", type=int) is None
            assert req.get("y", "20", type=int) == "20"

    def test_is_secure_returns_correct_value(self):
        with context("https://example.org/") as c:
            assert c.request.is_secure is True
        with context("http://example.org/") as c:
            assert c.request.is_secure is False

    def test_getint_returns_int(self):
        with context("http://example.org/?a=4") as c:
            assert c.request.getint("a") == 4

    def test_getint_raises_badrequest(self):
        with pytest.raises(exceptions.BadRequest):
            with context("http://example.org/") as c:
                c.request.getint("a", required=True)

        with pytest.raises(exceptions.BadRequest):
            with context("http://example.org/?a=four") as c:
                c.request.getint("a")

        with pytest.raises(exceptions.BadRequest):
            with context("http://example.org/?a=four") as c:
                assert c.request.getint("a", 0) == 0

    def test_getint_returns_default(self):
        with context("http://example.org/") as c:
            assert c.request.getint("a", 0) == 0

    def test_now(self):
        with context("http://example.org/") as c:
            now = c.request.now
            assert now.tzinfo is datetime.timezone.utc
            assert now is c.request.now


class TestPathEncoding(object):
    def test_url_is_quoted(self):
        with context(SCRIPT_NAME=fixtures.wsgi_unicode_path, PATH_INFO="") as c:
            assert c.request.url == "http://localhost" + fixtures.quoted_unicode_path

    def test_application_url_is_quoted(self):
        with context(SCRIPT_NAME=fixtures.wsgi_unicode_path, PATH_INFO="") as c:
            assert (
                c.request.application_url
                == "http://localhost" + fixtures.quoted_unicode_path
            )

    def test_parsed_url_is_quoted(self):
        with context(SCRIPT_NAME=fixtures.wsgi_unicode_path, PATH_INFO="") as c:
            assert c.request.parsed_url.path == fixtures.quoted_unicode_path

    def test_path_is_unquoted(self):
        with context(SCRIPT_NAME=fixtures.wsgi_unicode_path, PATH_INFO="") as c:
            assert c.request.path == fixtures.unquoted_unicode_path

    def test_script_name_is_unquoted(self):
        with context(SCRIPT_NAME=fixtures.wsgi_unicode_path, PATH_INFO="") as c:
            assert c.request.script_name == fixtures.unquoted_unicode_path

    def test_path_info_is_unquoted(self):
        with context(PATH_INFO=fixtures.wsgi_unicode_path) as c:
            assert c.request.path_info == fixtures.unquoted_unicode_path

    def test_invalid_path_info_encoding_raises_bad_request(self):
        with context(PATH_INFO=fixtures.misquoted_wsgi_unicode_path) as c:
            with pytest.raises(exceptions.BadRequest):
                c.request.path_info

    def test_invalid_script_name_encoding_raises_bad_request(self):
        with context(SCRIPT_NAME=fixtures.misquoted_wsgi_unicode_path) as c:
            with pytest.raises(exceptions.BadRequest):
                c.request.script_name


class TestMakeURL(object):
    def assert_equal_with_query(self, url1, url2):
        if "?" not in url1:
            assert url1 == url2
            return

        base1, _, q1 = url1.partition("?")
        base2, _, q2 = url2.partition("?")
        assert base1 == base2
        assert sorted(q1.split(";")) == sorted(q2.split(";"))

    def test_it_returns_request_url(self):
        with context(SCRIPT_NAME="/script", PATH_INFO="/pathinfo") as c:
            assert c.request.make_url() == "http://localhost/script/pathinfo"

    def test_it_doesnt_double_quote_request_url(self):
        with context(SCRIPT_NAME="/script name", PATH_INFO="/path info") as c:
            assert c.request.make_url() == "http://localhost/script%20name/path%20info"

    def test_it_doesnt_double_quote_supplied_path_info(self):
        with context() as c:
            assert c.request.make_url(PATH_INFO="/x y") == "http://localhost/x%20y"

    def test_can_replace_path(self):
        with context(SCRIPT_NAME="/script", PATH_INFO="/pathinfo") as c:
            assert c.request.make_url(path="/foo bar") == "http://localhost/foo%20bar"

    def test_it_joins_path(self):
        with context(SCRIPT_NAME="/script name", PATH_INFO="/path info/") as c:
            assert (
                c.request.make_url(path="a/b c")
                == "http://localhost/script%20name/path%20info/a/b%20c"
            )

    def test_query_not_included_by_default(self):
        with context(QUERY_STRING="query=foo") as c:
            assert c.request.make_url() == "http://localhost/"

    def test_query_dict(self):
        with context() as c:
            self.assert_equal_with_query(
                c.request.make_url(query={"a": 1, "b": "2 3"}),
                "http://localhost/?a=1&b=2+3",
            )

    def test_query_drops_items_with_None_value(self):
        with context() as c:
            self.assert_equal_with_query(
                c.request.make_url(query={"a": 1, "b": None}),
                "http://localhost/?a=1",
            )
            self.assert_equal_with_query(
                c.request.make_url(query=[("a", 1), ("b", None)]),
                "http://localhost/?a=1",
            )

    def test_query_kwargs(self):
        with context() as c:
            self.assert_equal_with_query(
                c.request.make_url(query={"a": 1}, b=2),
                "http://localhost/?a=1&b=2",
            )

    def test_unicode_path(self):
        e = b"\xc3\xa9".decode("utf8")  # e-acute
        with context() as c:
            assert c.request.make_url(path=e) == "http://localhost/%C3%A9"

    def test_unicode_path_info(self):
        e = b"\xc3\xa9".decode("utf8")  # e-acute
        with context() as c:
            assert c.request.make_url(PATH_INFO=e) == "http://localhost/%C3%A9"

    def test_unicode_query(self):
        e = b"\xc3\xa9".decode("utf8")  # e-acute
        with context() as c:
            assert c.request.make_url(query={"e": e}) == "http://localhost/?e=%C3%A9"

    def test_it_quotes_paths(self):
        with context() as c:
            assert c.request.make_url(path="/a b") == "http://localhost/a%20b"

    def test_replace_query_replaces_existing(self):
        with context(QUERY_STRING="a=1") as c:
            assert (
                c.request.make_url(query_replace={"a": "2"}) == "http://localhost/?a=2"
            )

    def test_replace_query_adds_new_value(self):
        with context(QUERY_STRING="a=1") as c:
            self.assert_equal_with_query(
                c.request.make_url(query_replace={"b": "2"}),
                "http://localhost/?a=1&b=2",
            )

    def test_add_query_doesnt_replace_existing(self):
        with context(QUERY_STRING="a=1") as c:
            self.assert_equal_with_query(
                c.request.make_url(query_add={"a": "2"}),
                "http://localhost/?a=1&a=2",
            )

    def test_add_query_works_on_specified_query(self):
        with context(QUERY_STRING="a=1") as c:
            assert (
                c.request.make_url(query="", query_add={"a": "2"})
                == "http://localhost/?a=2"
            )

    def test_add_query_does_not_mutate_request_query(self):
        with context(QUERY_STRING="a=1") as c:
            c.request.make_url(query_add={"a": "2"})
            assert list(c.request.query.allitems()) == [("a", "1")]


class TestResolveURL(object):
    def test_relative_path(self):
        with context(SCRIPT_NAME="/script", PATH_INFO="/pathinfo") as c:
            assert c.request.resolve_url("foo") == "http://localhost/script/foo"

    def test_absolute_path_unspecified_relative(self):
        with context(SCRIPT_NAME="/script", PATH_INFO="/pathinfo") as c:
            assert (
                c.request.resolve_url("/foo", relative="app")
                == "http://localhost/script/foo"
            )

    def test_absolute_path_app_relative(self):
        with context(SCRIPT_NAME="/script", PATH_INFO="/pathinfo") as c:
            assert (
                c.request.resolve_url("/foo", relative="app")
                == "http://localhost/script/foo"
            )

    def test_absolute_path_server_relative(self):
        with context(SCRIPT_NAME="/script", PATH_INFO="/pathinfo") as c:
            assert (
                c.request.resolve_url("/foo", relative="server")
                == "http://localhost/foo"
            )

    def test_ignores_server_port_if_host_header_present(self):
        with context(
            environ={"wsgi.url_scheme": "https", "SERVER_PORT": "81"},
            HTTP_HOST="localhost",
        ) as c:
            assert c.request.resolve_url("/foo") == "https://localhost/foo"


class TestCurrentRequest(object):
    def test_currentrequest_returns_current_request(self):
        from fresco import currentrequest

        with context() as c:
            assert currentrequest() is c.request

    def test_currentrequest_returns_None(self):
        from fresco import currentrequest

        assert currentrequest() is None


class TestMultipart(object):
    filename = "test.txt"
    filedata = "123456\n"
    boundary = "---------------------------1234"

    post_data = (
        (
            "--{boundary}\r\n"
            "Content-Disposition: form-data; "
            'name="uploaded_file"; filename="{filename}"\r\n'
            "Content-Type: text/plain\r\n"
            "\r\n"
            "{filedata}\r\n"
            "--{boundary}"
            "--\r\n"
        )
        .format(boundary=boundary, filedata=filedata, filename=filename)
        .encode("latin1")
    )

    request_args = {
        "wsgi_input": post_data,
        "REQUEST_METHOD": "POST",
        "CONTENT_TYPE": "multipart/form-data; boundary=" + boundary,
        "CONTENT_LENGTH": str(len(post_data)),
    }

    def test_files_populated(self):
        with FrescoApp().requestcontext(**self.request_args) as c:
            request = c.request

            assert len(request.files) == 1
            assert "uploaded_file" in request.files

    def test_file_content_available(self):
        with FrescoApp().requestcontext(**self.request_args) as c:
            request = c.request

            b = BytesIO()
            request.files["uploaded_file"].save(b)
        assert b.getvalue() == self.filedata.encode("latin1")

    def test_headers_available(self):
        with FrescoApp().requestcontext(**self.request_args) as c:
            request = c.request

            assert (
                request.files["uploaded_file"].headers["content-type"] == "text/plain"
            )
            assert request.files["uploaded_file"].filename == self.filename

    def test_quotes_in_input_names_are_decoded(self):
        """
        Field names in multipart form data must be decoded as RFC822
        quoted-strings
        """
        data = (
            b"----BOUNDARY\r\n"
            b'Content-Disposition: form-data; name="\\"qtext\\"";\r\n'
            b"\r\n"
            b"foo\r\n"
            b"----BOUNDARY"
            b"--\r\n"
        )

        env = {
            "wsgi_input": data,
            "REQUEST_METHOD": "POST",
            "CONTENT_TYPE": "multipart/form-data; boundary=--BOUNDARY",
            "CONTENT_LENGTH": str(len(data)),
        }

        with FrescoApp().requestcontext(**env) as c:
            assert list(c.request.form.items()) == [('"qtext"', "foo")]



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from fresco.core import FrescoApp
from fresco.util.urls import normpath, make_query
from fresco.util.urls import is_safe_url
from fresco.util.urls import add_query

# Greek letters as unicode strings (require multi-byte representation in UTF-8)
alpha = b"\xce\xb1".decode("utf8")
beta = b"\xce\xb2".decode("utf8")
gamma = b"\xce\xb3".decode("utf8")


class TestNormPath(object):
    def test_empty_string(selfself):
        assert normpath("") == ""

    def test_root_url(self):
        assert normpath("/") == "/"

    def test_condenses_consecutive_slashes(self):
        assert normpath("//") == "/"
        assert normpath("///") == "/"

    def test_remove_single_dot(self):
        assert normpath("/./") == "/"

    def test_double_dot_interpreted(self):
        assert normpath("/../") == "/"
        assert normpath("/foo/../") == "/"

    def test_triple_dot_preserved(self):
        assert normpath("/.../") == "/.../"

    def test_combined_patterns(self):
        assert normpath("/..//../") == "/"
        assert normpath("/hello/.//dolly//") == "/hello/dolly/"
        assert normpath("///hello/.//dolly//./..//.//sailor") == "/hello/sailor"

    def test_trailing_slash_preserved(self):
        assert normpath("/sliced/bread/") == "/sliced/bread/"


class TestMakeQuery(object):
    def test_make_query(self):
        assert sorted(make_query(a="1", b=2).split("&")) == ["a=1", "b=2"]
        assert make_query(a="one two three") == "a=one+two+three"
        assert make_query(a=["one", "two", "three"]) == "a=one&a=two&a=three"

    def test_make_query_unicode(self):
        assert (
            make_query(a=[alpha, beta, gamma], charset="utf8")
            == "a=%CE%B1&a=%CE%B2&a=%CE%B3"
        )

    def test_make_query_unicode_default_encoding(self):
        assert make_query(a=[alpha, beta, gamma], charset="utf8") == make_query(
            a=[alpha, beta, gamma]
        )


class TestSafeURL(object):
    def test_it_loads_defaults_from_context(self):
        with FrescoApp().requestcontext(
            **{"wsgi.url_scheme": "https", "HTTP_HOST": "foo.example.org"}
        ):
            assert is_safe_url("http://foo.example.org/") is True
            assert is_safe_url("https://foo.example.org/") is True
            assert is_safe_url("//foo.example.org/") is True
            assert is_safe_url("http://example.org/") is False
            assert is_safe_url("https://example.org/") is False
            assert is_safe_url("//example.org/") is False

    def test_it_passes_safe_urls(self):
        safe = [
            "/foo",
            "http://good.example.org/x",
            "https://good.example.org/",
            "HTTPS://good.example.org/",
            "//good.example.org/",
            "/path?next_url=http://bad.example.org/",
        ]
        for u in safe:
            assert is_safe_url(u, allowed_hosts={"good.example.org"}) is True

    def test_it_handles_port_numbers_correctly(self):
        def test(scheme, server_name="localhost", server_port="80", http_host=None):
            env = {
                "SERVER_NAME": server_name,
                "SERVER_PORT": server_port,
                "wsgi.url_scheme": scheme,
                "HTTP_HOST": http_host,
            }

            def test(url, expected):
                with FrescoApp().requestcontext(**env):
                    assert is_safe_url(url) is expected

            return test

        # server_name, server_port set, but no http_host
        test("http", "example", "80")("http://example/", True)
        test("http", "example", "80")("http://example:80/", True)
        test("http", "example", "80")("https://example/", True)
        test("http", "example", "80")("https://example:443/", True)
        test("http", "example", "80")("http://example:443/", False)
        test("http", "example", "80")("http://example:8080/", False)

        test("https", "example", "443")("http://example/", True)
        test("https", "example", "443")("http://example:80/", True)
        test("https", "example", "443")("https://example/", True)
        test("https", "example", "443")("https://example:443/", True)
        test("https", "example", "443")("https://example:80/", False)

        test("http", "example", "8080")("http://example:8080/", True)
        test("http", "example", "8080")("https://example:8080/", True)
        test("http", "example", "8080")("http://example/", False)
        test("http", "example", "8080")("http://example:80/", False)

        # http_host set - strict match on origin
        test("http", http_host="example")("http://example/", True)
        test("http", http_host="example")("http://example:80/", True)
        test("http", http_host="example")("https://example/", True)
        test("http", http_host="example")("https://example:443/", True)
        test("http", http_host="example")("http://example:443/", False)
        test("http", http_host="example")("http://example:8080/", False)

        test("https", http_host="example")("http://example/", True)
        test("https", http_host="example")("http://example:80/", True)
        test("https", http_host="example")("https://example/", True)
        test("https", http_host="example")("https://example:443/", True)
        test("https", http_host="example")("http://example:443/", False)
        test("https", http_host="example")("http://example:8080/", False)

        test("http", http_host="example:80")("http://example/", True)
        test("http", http_host="example:80")("http://example:80/", True)
        test("http", http_host="example:80")("https://example/", True)
        test("http", http_host="example:80")("https://example:443/", True)
        test("http", http_host="example:80")("http://example:443/", False)
        test("http", http_host="example:80")("http://example:8080/", False)

        test("https", http_host="example:443")("http://example/", True)
        test("https", http_host="example:443")("http://example:80/", True)
        test("https", http_host="example:443")("https://example/", True)
        test("https", http_host="example:443")("https://example:443/", True)
        test("https", http_host="example:443")("http://example:443/", False)
        test("https", http_host="example:443")("http://example:8080/", False)

        test("http", http_host="example:8080")("http://example:8080/", True)
        test("http", http_host="example:8080")("https://example:8080/", True)
        test("http", http_host="example:8080")("http://example/", False)
        test("http", http_host="example:8080")("http://example:80/", False)

    def test_it_catches_unsafe_urls(self):
        unsafe = [
            "///foo",
            "http://good.example.org@bad.example.org/",
            "http://good.example.org:good.example.org@bad.example.org/",
            "javascript:do_something_bad()",
            "",
            "\x00http://good.example.org",
            "http://good.example.org\nHost: http://bad.example.org",
        ]
        for u in unsafe:
            assert is_safe_url(u, allowed_hosts={"good.example.org"}) is False


class TestAddQuery:

    def test_it_adds_query(self):
        url = add_query("http://localhost", {"foo": "bar"})
        assert url == "http://localhost?foo=bar"

    def test_it_adds_query_from_list(self):
        url = add_query("http://localhost", [("foo", "bar")])
        assert url == "http://localhost?foo=bar"

    def test_it_adds_query_from_keywords(self):
        url = add_query("http://localhost", foo="bar")
        assert url == "http://localhost?foo=bar"

    def test_it_appends_query(self):
        url = add_query("http://localhost?foo=bar", {"foo": "baz"})
        assert url == "http://localhost?foo=bar&foo=baz"



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
import pytest

from fresco.util.security import check_equal_constant_time


class TestSecurity(object):
    def test_check_equal_constant_time_returns_equality(self):
        assert check_equal_constant_time("", "") is True
        assert check_equal_constant_time("abcabcabc", "abcabcabc") is True

    def test_check_equal_constant_time_returns_inequality(self):
        assert check_equal_constant_time(" ", "") is False
        assert check_equal_constant_time("abcabcabc", "abcabcabx") is False
        assert check_equal_constant_time("abcabcabc", "abcabcabx") is False
        assert check_equal_constant_time("abcabcabc", "abcabcabde") is False
        assert check_equal_constant_time("abcabcabc", "abcabcab") is False

    def test_check_equal_constant_time_raises_an_error_on_non_strings(self):
        with pytest.raises(TypeError):
            check_equal_constant_time(None, None)



from collections.abc import Mapping
from collections.abc import Iterable
from collections.abc import Callable
from types import TracebackType
from typing import Any
from typing import Optional
import typing as t

import fresco.response # noqa


QuerySpec = t.Union[
    Mapping[str, Any],
    Iterable[tuple[str, Any]]
]
ViewCallable = Callable[..., "fresco.response.Response"]
ExcInfo = tuple[type[BaseException], BaseException, TracebackType]
OptionalExcInfo = Optional[ExcInfo]


HeaderList = list[tuple[str, str]]
HeadersList = HeaderList
WSGIEnviron = dict[str, Any]
WriteCallable = Callable[[bytes], object]
StartResponse = Callable[[str, HeaderList, OptionalExcInfo], WriteCallable]
WSGIApplication = Callable[
    [
        WSGIEnviron,
        StartResponse,
    ],
    Iterable[bytes]
]



# encoding=UTF-8
# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from unittest.mock import Mock
import pytest

from fresco import FrescoApp
from fresco.cookie import Cookie
from fresco.response import Response
from fresco.routing import GET
from fresco.routing import RouteNotFound


class TestCannedResponses(object):
    def test_redirect(self):
        with FrescoApp().requestcontext():
            r = Response.redirect("/new_location")
        assert b"http://localhost/new_location" in b"".join(r.content_iterator)
        assert r.get_header("Location") == "http://localhost/new_location"
        assert r.status == "302 Found"

    def test_redirect_temporary(self):
        with FrescoApp().requestcontext():
            r = Response.redirect_temporary("/new_location")
        assert b"http://localhost/new_location" in b"".join(r.content_iterator)
        assert r.get_header("Location") == "http://localhost/new_location"
        assert r.status == "302 Found"

    def test_redirect_permanent(self):
        with FrescoApp().requestcontext():
            r = Response.redirect_permanent("/new_location")
        assert b"http://localhost/new_location" in b"".join(r.content_iterator)
        assert r.get_header("Location") == "http://localhost/new_location"
        assert r.status == "301 Moved Permanently"

    def test_redirect_to_view(self):
        app = FrescoApp()
        view = Mock(return_value=Response())
        app.route("/arfle/<barfle:int>", GET, view)
        with app.requestcontext():
            r = Response.redirect(view, barfle=42)
        assert r.get_header("Location") == "http://localhost/arfle/42"

    def test_redirect_to_view_spec(self):
        app = FrescoApp()
        view = Mock(return_value=Response())
        app.route("/arfle/<barfle:int>", GET, view, name="gloop")
        with app.requestcontext():
            r = Response.redirect("gloop", barfle=42)
        assert r.get_header("Location") == "http://localhost/arfle/42"

    def test_redirect_to_simple_string(self):
        with FrescoApp().requestcontext():
            r = Response.redirect("gloop")
        assert r.get_header("Location") == "http://localhost/gloop"

    def test_redirect_raises_RouteNotFound(self):
        view = Mock()
        with FrescoApp().requestcontext():
            with pytest.raises(RouteNotFound):
                Response.redirect("gloop", arfle="a")
            with pytest.raises(RouteNotFound):
                Response.redirect(view)

    def test_redirect_to_unsafe_url(self):
        with pytest.raises(ValueError):
            with FrescoApp().requestcontext():
                Response.redirect("http://bad.example.com/")

    def test_redirect_uses_fallback_for_unsafe_url(self):
        app = FrescoApp()
        view = Mock(return_value=Response())
        app.route("/arfle/<barfle:int>", GET, view, name="gloop")
        with app.requestcontext():
            r = Response.redirect("http://bad.example.com/", "gloop", barfle=23)
        assert r.get_header("Location") == "http://localhost/arfle/23"

    def test_not_found(self):
        with FrescoApp().requestcontext():
            r = Response.not_found()
        assert b"not found" in b"".join(r.content_iterator).lower()
        assert r.status_code == 404

    def test_error(self):
        with FrescoApp().requestcontext():
            r = Response.error()
        assert b"internal server error" in b"".join(r.content_iterator).lower()
        assert r.status_code == 500

    def test_method_not_allowed(self):
        with FrescoApp().requestcontext():
            r = Response.method_not_allowed(valid_methods=("PUT", "DELETE"))
        assert b"not allowed" in b"".join(r.content_iterator).lower()
        assert r.status == "405 Method Not Allowed"
        assert r.get_header("Allow") == "PUT,DELETE"

    def test_unauthorized(self):
        with FrescoApp().requestcontext():
            r = Response.unauthorized(authenticate='myscheme realm="spaghetti"')
        assert r.status == "401 Unauthorized"
        assert r.get_header("WWW-Authenticate") == 'myscheme realm="spaghetti"'

    def test_unauthorized_basic(self):
        with FrescoApp().requestcontext():
            r = Response.unauthorized_basic(realm="spaghetti")
        assert r.status == "401 Unauthorized"
        assert r.get_header("WWW-Authenticate") == 'Basic realm="spaghetti"'

    def test_meta_refresh(self):
        with FrescoApp().requestcontext():
            r = Response.meta_refresh("/next_page")
        content = b"".join(r.content_iterator)
        assert b'<a href="http://localhost/next_page">' in content
        assert (
            b'<meta http-equiv="refresh" content="0;'
            b' url=http://localhost/next_page">' in content
        )
        assert r.status == "200 OK"
        assert r.get_header("Content-Type") == "text/html"

    def test_json(self):
        with FrescoApp().requestcontext():
            r = Response.json(
                {"arfle": "barfle"},
                status=201,
                indent=4,
                separators=(",", ":\t"),
            )
        assert r.status == "201 Created"
        assert r.get_header("Content-Type") == "application/json"
        assert b"".join(r.content_iterator) == b'{\n    "arfle":\t"barfle"\n}'


class TestResponse(object):
    def test_unknown_response_header_from_python_symbol(self):
        r = Response(x_myheader="Foo", content_type=None)
        assert r.headers == [("X-Myheader", "Foo")]

    def test_standard_headers_capitalized_from_symbol(self):
        r = Response(etag="foo")
        assert r.headers == [("ETag", "foo")]

        r = Response(x_ua_compatible="foo")
        assert r.headers == [("X-UA-Compatible", "foo")]

    def test_content_type_added_only_if_content(self):
        r = Response(content=[])
        assert "Content-Type" in dict(r.headers)

        r = Response()
        assert "Content-Type" not in dict(r.headers)

        r = Response(content=None)
        assert "Content-Type" not in dict(r.headers)

    def test_add_response_header(self):
        r = Response(x_myheader="Foo")
        r = r.add_header("X-Myheader", "Bar")
        assert r.headers == [("X-Myheader", "Foo"), ("X-Myheader", "Bar")]

    def test_set_content_type_header(self):
        r = Response(content_type="text/rtf; charset=ISO-8859-1")
        assert r.get_header("Content-Type") == "text/rtf; charset=ISO-8859-1"

    def test_set_cookie(self):
        r = Response(set_cookie=Cookie(name="fruit", value="banana"))
        assert r.get_header("Set-Cookie") == "fruit=banana;Path=/;SameSite=Lax"

    def test_delete_cookie(self):
        r = Response().delete_cookie(name="fruit")
        assert r.get_header("Set-Cookie") == (
            "fruit=;Expires=Thu, 01 Jan 1970 00:00:00 GMT;Max-Age=0;Path=/;SameSite=Lax"
        )

    def test_response_get_headers(self):
        r = Response(
            ["whoa nelly!"],
            content_type="text/plain",
            x_test_header=["1", "2"],
        )
        assert r.get_header("content-type") == "text/plain"
        assert r.get_header("x-test-header") == "1,2"
        assert r.get_header("X-Test-Header") == "1,2"
        assert r.get_header("x-does-not-exist", "boo!") == "boo!"

    def test_response_buffered_sets_content_length(self):
        r = Response(["whoa nelly!"]).buffered()
        assert r.get_header("content-length") == "11"

        r = Response(["whoa nélly!"]).buffered()
        assert r.get_header("content-length") == "12"

    def test_it_encodes_a_unicode_string_according_to_content_type(self):
        r = Response("café", content_type="text/plain; charset=utf-8")
        assert list(r({}, Mock())) == ["café".encode("utf-8")]
        r = Response("café", content_type="text/plain; charset=latin-1")
        assert list(r({}, Mock())) == ["café".encode("latin-1")]


class TestAddVary(object):
    def test_it_adds_a_vary_header(self):
        r = Response(content_type="text/plain").add_vary("Accept-Encoding")
        assert ("Vary", "Accept-Encoding") in r.headers

    def test_it_extends_an_existing_header(self):
        r = (
            Response(content_type="text/plain")
            .add_vary("Accept-Encoding")
            .add_vary("Cookie")
        )
        assert "Cookie" in r.get_header("vary")
        assert "Accept-Encoding" in r.get_header("vary")



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
import sys
from functools import wraps
import typing as t

from fresco import Response

_marker = object()


def onerror(
    exceptions: t.Union[t.Type[Exception], tuple[t.Type[Exception], ...]], handler
) -> t.Callable:
    """
    Return a decorator that can replace or update the return value of the
    function if an exception is raised
    """

    try:
        if isinstance(exceptions, type):
            exceptions = (exceptions,)
    except TypeError:
        pass

    def decorator(func):
        @wraps(func)
        def decorated(*fargs, **fkwargs):
            try:
                return func(*fargs, **fkwargs)
            except exceptions:
                exc_info = sys.exc_info()
                return handler(exc_info, *fargs, **fkwargs)

        return decorated

    return decorator


def json_response(
    data=_marker,
    indent=None,
    separators=(",", ":"),
    content_type="application/json",
    **kwargs
):
    """
    JSON encode the function's result and return a response object with the
    content-type ``application/json``.

    May also be used as a regular function to directly create a JSON encoded
    response, however it's preferrable to use
    :meth:`~fresco.response.Response.json` for this.

    :param data: The data to json encode.
                 If left unspecified, ``json_response``
                 will act as a function decorator.
    :param indent: The indent level. Defaults to ``None`` (no pretty printing)
    :param separators: Defaults to ``(',', ':')`` for the most compact JSON
                       representation
    :param kwargs: Other keyword arguments are passed to ``json.dumps``. These
                   may be used to change encoding paramters, for example
                   overriding the default ``JSONEncoder`` class.


    """
    # Called as a decorator or decorator factory
    if data is _marker or callable(data):

        def json_response_decorator(func):
            @wraps(func)
            def json_response_decorated(*fa, **fkw):
                return Response.json(func(*fa, **fkw), indent, separators, **kwargs)

            return json_response_decorated

        # Called as a decorator factory (``@json_response()``)
        if data is _marker:
            return json_response_decorator

        # Called as a decorator (ie no parens: ``@json_response``)
        return json_response_decorator(data)

    # Called as a regular function (or via ``Route(...).filter``)
    else:
        if isinstance(data, Response):
            return data

        return Response.json(
            data,
            indent=indent,
            separators=separators,
            content_type=content_type,
            **kwargs
        )



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from fresco import FrescoApp
from fresco import context
from fresco.middleware import XForwarded
from fresco.response import Response
from fresco.routing import GET


class TestXForwarded(object):
    def get_app(self, *args, **kwargs):
        app = FrescoApp()

        @app.route("/", GET)
        def view():
            request = context.request
            return Response([request.url, request.environ["REMOTE_ADDR"]])

        app.add_middleware(XForwarded, *args, **kwargs)
        return app

    def test_forwards_x_forwarded_for(self):
        app = self.get_app()
        with app.requestcontext(
            "/", REMOTE_ADDR="127.0.0.1", HTTP_X_FORWARDED_FOR="1.2.3.4"
        ):
            url, addr = app.view().content_iterator
            assert addr == b"1.2.3.4"

    def test_forwards_x_forwarded_host(self):
        app = self.get_app()
        with app.requestcontext(
            "/",
            REMOTE_ADDR="127.0.0.1",
            HTTP_X_FORWARDED_HOST="frontendserver",
        ):
            url, addr = app.view().content_iterator
        assert url == b"http://frontendserver/"

    def test_forwards_x_forwarded_ssl(self):
        app = self.get_app()
        with app.requestcontext(
            "/",
            REMOTE_ADDR="127.0.0.1",
            HTTP_X_FORWARDED_SSL="on",
            HTTP_X_FORWARDED_HOST="localhost",
        ):
            url, addr = app.view().content_iterator
            assert url == b"https://localhost/"

    def test_forwards_x_forwarded_proto(self):
        app = self.get_app()
        with app.requestcontext(
            "/",
            REMOTE_ADDR="127.0.0.1",
            HTTP_X_FORWARDED_PROTO="https",
            HTTP_X_FORWARDED_HOST="localhost",
        ):
            url, addr = app.view().content_iterator
            assert url == b"https://localhost/"

    def test_reports_first_trusted_ip(self):
        app = self.get_app(trusted=["127.0.0.1", "3.3.3.3"])
        with app.requestcontext(
            "/",
            REMOTE_ADDR="127.0.0.1",
            HTTP_X_FORWARDED_FOR="1.1.1.1, " "2.2.2.2, 3.3.3.3",
        ):
            url, addr = app.view().content_iterator
            assert addr == b"2.2.2.2"

    def test_reports_first_listed_ip_as_remote_addr_when_untrusted(self):
        app = self.get_app()
        with app.requestcontext(
            "/",
            REMOTE_ADDR="127.0.0.1",
            HTTP_X_FORWARDED_FOR="1.1.1.1, 2.2.2.2",
        ):
            url, addr = app.view().content_iterator
            assert addr == b"1.1.1.1"



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
import pytest
from fresco.multidict import MultiDict


def is_consistent(multidict: MultiDict) -> bool:
    "Verify that the MultiDict has internal consistency"
    dictallitems = sorted((k, v) for k, vs in multidict._dict.items() for v in vs)
    orderallitems = sorted(multidict._order)
    return list(sorted(dictallitems)) == list(sorted(orderallitems))


class TestMultDict(object):
    def test_getitem_returns_only_first(self):
        m = MultiDict([("a", 1), ("a", 2), ("b", 3)])
        assert m["a"] == 1

    def test_getitem_throws_keyerror(self):
        m = MultiDict([("a", 1), ("a", 2), ("b", 3)])
        with pytest.raises(KeyError):
            m["c"]

    def test_get_returns_only_first(self):
        m = MultiDict([("a", 1), ("a", 2), ("b", 3)])
        assert m.get("a") == 1

    def test_get_returns_default(self):
        m = MultiDict()
        assert m.get("a") is None
        assert m.get("a", "foo") == "foo"

    def test_getlist_returns_list(self):
        m = MultiDict([("a", 1), ("a", 2)])
        assert m.getlist("a") == [1, 2]
        assert m.getlist("b") == []

    def test_copy_is_equal(self):
        m = MultiDict([("a", 1), ("a", 2)])
        assert list(m.allitems()) == list(m.copy().allitems())

    def test_copy_is_independent(self):
        m = MultiDict([("a", 1), ("a", 2)])
        n = m.copy()
        n["b"] = "foo"
        assert list(m.allitems()) == [("a", 1), ("a", 2)]
        assert list(n.allitems()) == [("a", 1), ("a", 2), ("b", "foo")]

    def test_fromkeys(self):
        m = MultiDict.fromkeys(["a", "b"])
        assert list(m.allitems()) == [("a", None), ("b", None)]

    def test_fromkeys_with_value(self):
        m = MultiDict.fromkeys(["a", "b"], 42)
        assert list(m.allitems()) == [("a", 42), ("b", 42)]

    def test_items_only_returns_first_of_each(self):
        m = MultiDict([("a", 1), ("a", 2)])
        assert list(m.items()) == [("a", 1)]

    def test_listitems_returns_lists(self):
        m = MultiDict([("a", 1), ("a", 2)])
        assert list(m.listitems()) == [("a", [1, 2])]

    def test_allitems_returns_all(self):
        m = MultiDict([("a", 1), ("a", 2)])
        assert list(m.allitems()) == [("a", 1), ("a", 2)]

    def test_allitems_returns_iterator(self):
        m = MultiDict([("a", 1), ("a", 2)])
        i = m.allitems()
        assert next(i) == ("a", 1)
        assert next(i) == ("a", 2)
        with pytest.raises(StopIteration):
            next(i)

    def test_keys(self):
        m = MultiDict([("a", 1), ("a", 2), ("b", 3)])
        i = iter(m.keys())
        assert next(i) == "a"
        assert next(i) == "b"
        with pytest.raises(StopIteration):
            next(i)

    def test_values(self):
        m = MultiDict([("a", 1), ("a", 2), ("b", 3)])
        i = iter(m.values())
        assert next(i) == 1
        assert next(i) == 3
        with pytest.raises(StopIteration):
            next(i)

    def test_listvalues(self):
        m = MultiDict([("a", 1), ("a", 2), ("b", 3)])
        assert list(m.listvalues()) == [[1, 2], [3]]

    def test_setitem(self):
        m = MultiDict()
        m["a"] = 1
        assert m["a"] == 1
        assert m.getlist("a") == [1]
        assert is_consistent(m)

    def test_setitem_replaces_existing(self):
        m = MultiDict()
        m["a"] = 1
        m["a"] = 2
        assert m.getlist("a") == [2]
        assert is_consistent(m)

    def test_delitem(self):
        m = MultiDict([("a", 1), ("a", 2), ("b", 3)])

        del m["a"]
        assert list(m.allitems()) == [("b", 3)]
        assert is_consistent(m)

        m = MultiDict([("a", 1), ("a", 2), ("b", 3)])
        del m["b"]
        assert list(m.allitems()) == [("a", 1), ("a", 2)]
        assert is_consistent(m)

    def test_iterable(self):
        m = MultiDict([("a", 1), ("a", 2), ("b", 3)])
        assert list(sorted(iter(m))) == ["a", "b"]

    def test_update_with_list(self):
        m = MultiDict([("a", 1), ("b", 2)])
        m.update([("a", 2), ("x", 2)])
        assert list(sorted(m.allitems())) == [("a", 2), ("b", 2), ("x", 2)]
        assert is_consistent(m)

    def test_update_with_dict(self):
        m = MultiDict([("a", 1), ("b", 2)])
        m.update({"a": 2, "x": 2})
        assert list(sorted(m.allitems())) == [("a", 2), ("b", 2), ("x", 2)]
        assert is_consistent(m)

    def test_update_with_multidict(self):
        m = MultiDict([("a", 1), ("b", 2)])
        m.update(MultiDict([("a", 2), ("x", 2)]))
        assert list(sorted(m.allitems())) == [("a", 2), ("b", 2), ("x", 2)]
        assert is_consistent(m)

    def test_update_with_kwargs(self):
        m = MultiDict([("a", 1), ("b", 2)])
        m.update(a=2, x=2)
        assert sorted(list(m.allitems())) == [("a", 2), ("b", 2), ("x", 2)]
        assert is_consistent(m)

    def test_extend_with_list(self):
        m = MultiDict([("a", 1), ("b", 2)])
        m.extend([("a", 2), ("x", 2)])
        assert list(m.allitems()) == [("a", 1), ("b", 2), ("a", 2), ("x", 2)]
        assert is_consistent(m)

    def test_extend_with_dict(self):
        m = MultiDict([("a", 1), ("b", 2)])
        m.extend({"a": 2, "x": 2})
        assert list(sorted(m.allitems())) == [
            ("a", 1),
            ("a", 2),
            ("b", 2),
            ("x", 2),
        ]
        assert is_consistent(m)

    def test_extend_with_multidict(self):
        m = MultiDict([("a", 1), ("b", 2)])
        m.extend(MultiDict([("a", 2), ("x", 2)]))
        assert list(sorted(m.allitems())) == [
            ("a", 1),
            ("a", 2),
            ("b", 2),
            ("x", 2),
        ]
        assert is_consistent(m)

    def test_extend_with_kwargs(self):
        m = MultiDict([("a", 1), ("b", 2)])
        m.extend(a=2, x=2)
        assert list(sorted(m.allitems())) == [
            ("a", 1),
            ("a", 2),
            ("b", 2),
            ("x", 2),
        ]
        assert is_consistent(m)

    def test_pop(self):
        m = MultiDict([("a", 1), ("b", 2), ("a", 3)])
        assert m.pop("b") == 2
        assert list(sorted(m.allitems())) == [("a", 1), ("a", 3)]
        assert is_consistent(m)

    def test_pop_returns_first_item(self):
        m = MultiDict([("a", 1), ("b", 2), ("a", 3)])
        assert m.pop("a") == 1
        assert list(sorted(m.allitems())) == [("a", 3), ("b", 2)]
        assert is_consistent(m)

    def test_pop_with_default(self):
        m = MultiDict([])
        assert m.pop("c", "foo") == "foo"
        assert is_consistent(m)

    def test_popitem(self):
        m = MultiDict([("a", 1), ("b", 2), ("b", 3)])

        assert m.popitem() == ("b", 2)
        assert is_consistent(m)
        assert m.popitem() == ("b", 3)
        assert is_consistent(m)
        assert m.popitem() == ("a", 1)
        assert is_consistent(m)

        with pytest.raises(KeyError):
            m.popitem()

    def test_len(self):
        m = MultiDict([("a", 1), ("b", 2), ("b", 3)])
        assert len(m) == 2



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from decimal import Decimal
from tempfile import NamedTemporaryFile
from tempfile import TemporaryDirectory
from unittest.mock import Mock
import contextlib
import os
import pathlib
import sys

import pytest

from fresco.options import Options
from fresco.options import override_options
from fresco.options import parse_key_value_pairs
from fresco.options import dict_from_options


class TestOptions(object):
    def test_options_dictionary_access(self):
        options = Options()
        options["x"] = 1
        assert options["x"] == 1

    def test_options_attribute_access(self):
        options = Options()
        options.x = 1
        assert options.x == 1

    def test_options_raises_AttributeError(self):
        with pytest.raises(AttributeError):
            Options().x

    def test_options_works_with_getattr(self):
        assert getattr(Options(), "x", 1) == 1
        assert getattr(Options({"x": 2}), "x", 1) == 2
        with pytest.raises(AttributeError):
            assert getattr(Options(), "x")

    def test_options_update_from_object(self):
        class Foo:
            a = 1
            b = 2

        options = Options()
        options.update_from_object(Foo())
        assert options == {"a": 1, "b": 2}

    def test_options_update_from_object_loads_underscore_names(self):
        class Foo:
            pass

        options = Options()
        options.update_from_object(Foo(), True)
        assert "__module__" in options

    def test_options_update_from_file(self):
        with NamedTemporaryFile() as tmpfile:
            tmpfile.write(b"a = 1\nb = 2\n")
            tmpfile.flush()

            options = Options()
            options.update_from_file(tmpfile.name)
            assert options == {"a": 1, "b": 2}

    def test_options_update_from_file_has_dunder_file_global(self):
        with NamedTemporaryFile() as tmpfile:
            tmpfile.write(b"a = __file__")
            tmpfile.flush()

            options = Options()
            options.update_from_file(tmpfile.name)
            assert options == {"a": tmpfile.name}

    def test_options_respects_all(self):
        with NamedTemporaryFile() as tmpfile:
            tmpfile.write(b"__all__ = ['a']\n" b"a = 1\n" b"b = 2\n")
            tmpfile.flush()

            options = Options()
            options.update_from_file(tmpfile.name)
            assert options == {"a": 1}

    def test_update_from_file_doesnt_add_module(self):
        with NamedTemporaryFile() as tmpfile:
            options = Options()
            saved_modules = list(sorted(sys.modules.keys()))
            options.update_from_file(tmpfile.name)
            assert list(sorted(sys.modules.keys())) == saved_modules

    def test_options_copy_returns_options(self):
        assert isinstance(Options().copy(), Options)


class TestOverrideOptions:
    def test_override_options_with_object(self):
        options = Options(foo=1)
        with override_options(options, {"foo": 2, "bar": "a"}):
            assert options["foo"] == 2
            assert options["bar"] == "a"
        assert options["foo"] == 1
        assert "bar" not in options

    def test_override_options_with_kwargs(self):
        options = Options(foo=1)
        with override_options(options, foo=2, bar="a"):
            assert options["foo"] == 2
            assert options["bar"] == "a"
        assert options["foo"] == 1
        assert "bar" not in options


class TestLoadKeyValuePairs:
    def test_it_loads_strings(self):
        assert parse_key_value_pairs({}, ["a=b"]) == {"a": "b"}

    def test_it_loads_ints(self):
        assert parse_key_value_pairs({}, ["a=100"]) == {"a": 100}
        assert parse_key_value_pairs({}, ["a=-100"]) == {"a": -100}
        assert parse_key_value_pairs({}, ["a=+100"]) == {"a": 100}
        # leading zero - not treated as an int
        assert parse_key_value_pairs({}, ["a=01"]) == {"a": "01"}

    def test_it_loads_bools(self):
        assert parse_key_value_pairs({}, ["a=true"]) == {"a": True}
        assert parse_key_value_pairs({}, ["a=True"]) == {"a": True}
        assert parse_key_value_pairs({}, ["a=TRUE"]) == {"a": True}
        assert parse_key_value_pairs({}, ["a=false"]) == {"a": False}
        assert parse_key_value_pairs({}, ["a=False"]) == {"a": False}
        assert parse_key_value_pairs({}, ["a=FALSE"]) == {"a": False}

    def test_it_loads_decimals(self):
        assert parse_key_value_pairs({}, ["a=1.2"]) == {"a": Decimal("1.2")}
        assert parse_key_value_pairs({}, ["a=+1."]) == {"a": Decimal("1")}
        assert parse_key_value_pairs({}, ["a=-.1"]) == {"a": Decimal("-0.1")}

    def test_it_ignores_comments(self):
        assert parse_key_value_pairs({}, ["a=1", "b=2 #comment", "#c=3"]) == {
            "a": 1,
            "b": 2,
        }

    def test_it_interpolates(self):
        assert parse_key_value_pairs({}, ["a=${x}"]) == {"a": "${x}"}
        assert parse_key_value_pairs({}, ["a=$x"]) == {"a": "$x"}
        assert parse_key_value_pairs({"x": 1}, ["a=${x}"]) == {"a": 1}
        assert parse_key_value_pairs({"x": 1}, ["a=$x"]) == {"a": 1}
        assert parse_key_value_pairs({}, ["a=1", "b=$a"]) == {"a": 1, "b": 1}


class TestLoadOptions:
    def check_loadoptions(self, tmpdir, files, sources="*", tags=[], expected={}):
        """
        Write the files indicated in ``sources`` to the given temporary directory,

        Create an Options object and populate it from the specified sources/tags.

        Assert that the loaded options is equal to the value of ``expected``.
        """
        t = pathlib.Path(tmpdir)
        for fname in files:
            with (t / fname).open("w", encoding="UTF-8") as f:
                f.write(files[fname])

        @contextlib.contextmanager
        def optionsdir():
            def loadopts(sources="*", tags=[], strict=False, **kw):
                return Options().load(sources, tags, strict=False, **kw)

            saved = os.getcwd()
            os.chdir(t)
            try:
                yield loadopts
            finally:
                os.chdir(saved)

        if expected:
            with optionsdir() as loadopts:
                assert loadopts(sources, tags) == expected
        else:
            return optionsdir()

    def test_it_loads_kvp_files(self, tmpdir):
        self.check_loadoptions(tmpdir, {"a": "x = 2"}, expected={"x": 2})
        self.check_loadoptions(
            tmpdir, {"a": "x = 2\ny = ${x}"}, expected={"x": 2, "y": 2}
        )
        with self.check_loadoptions(tmpdir, {"a": "x = $__FILE__\ny=${x}"}) as loadopts:
            result = loadopts()
            assert result["x"] == result["y"] == str(pathlib.Path(tmpdir) / "a")

    def test_it_loads_json(self, tmpdir):
        self.check_loadoptions(
            tmpdir, {"a.json": '{"a": ["b"]}'}, expected={"a": ["b"]}
        )

    def test_it_loads_py_files(self, tmpdir):
        self.check_loadoptions(tmpdir, {"a.py": "x = 2 * 2"}, expected={"x": 4})

    def test_py_files_have_options_in_namespace(self, tmpdir):
        self.check_loadoptions(
            tmpdir, {"a.py": "options['foo'] = 'bar'"}, expected={"foo": "bar"}
        )

    def test_it_selects_by_tag(self, tmpdir):
        with self.check_loadoptions(
            tmpdir,
            {
                "a.dev.txt": "a = 1",
                "a.staging.txt": "b = 1",
                "a.staging.local.txt": "c = 1",
                "a.dev.local.txt": "d = 1",
                "a.local.txt": "e = 1",
            },
        ) as loadopts:
            assert loadopts("*", ["dev"]) == {"a": 1}
            assert loadopts("*", ["dev", "local"]) == {"a": 1, "d": 1, "e": 1}
            assert loadopts("*", ["staging"]) == {"b": 1}
            assert loadopts("*", ["staging", "local"]) == {
                "b": 1,
                "c": 1,
                "e": 1,
            }
            assert loadopts("*", ["local"]) == {"e": 1}

    def test_it_loads_in_tag_order(self, tmpdir):
        with self.check_loadoptions(
            tmpdir,
            {
                "a": "a = 'a'",
                "a.0-dev.txt": "a = ${a} a.dev",
                "a.local.txt": "a = ${a} a.local",
                "b.dev.txt": "a = ${a} b.dev",
            },
        ) as loadopts:
            assert loadopts("*", ["dev", "local"]) == {"a": "a a.dev b.dev a.local"}
            assert loadopts("*", ["local", "dev"]) == {"a": "a a.local a.dev b.dev"}

    def test_it_loads_in_priority_order(self, tmpdir):
        with self.check_loadoptions(
            tmpdir,
            {
                "a": "a = 'a'",
                "a.100-dev.txt": "a = ${a} a.100-dev",
                "a.local.txt": "a = ${a} a.local",
                "b.dev.txt": "a = ${a} b.dev",
            },
        ) as loadopts:
            assert loadopts("*", ["dev", "local"]) == {"a": "a b.dev a.local a.100-dev"}
            assert loadopts("*", ["local", "dev"]) == {"a": "a a.local b.dev a.100-dev"}

    def test_it_loads_in_priority_order_without_tags(self, tmpdir):
        with self.check_loadoptions(
            tmpdir,
            {
                "a": "a = a",
                "b.100": "a = ${a} b",
                "a.200.txt": "a = ${a} 100",
            },
        ) as loadopts:
            assert loadopts("*") == {"a": "a b 100"}

    def test_it_loads_from_os_environ(self, tmpdir):
        with setenv(a="2"):
            with self.check_loadoptions(tmpdir, {"a.txt": "a = 1"}) as loadopts:
                assert loadopts("*", [], use_environ=False) == {"a": 1}
                assert loadopts("*", [], use_environ=True) == {"a": 2}

    def test_it_calls_callbacks(self, tmpdir):
        with self.check_loadoptions(tmpdir, {"a.txt": "a = 1"}):
            options = Options()
            mock = Mock()
            options.onload(mock)
            options.load(str(pathlib.Path(tmpdir) / "*"))
            assert mock.called

    def test_it_sets_directory(self, tmpdir):
        with self.check_loadoptions(tmpdir, {"a.txt": "a = 1"}) as loadopts:
            with TemporaryDirectory() as tmpdir2:
                os.chdir(tmpdir2)
                assert loadopts("*") == {}
                assert loadopts("*", dir=tmpdir) == {"a": 1}

    def test_it_accepts_a_list_of_filespecs(self, tmpdir):
        self.check_loadoptions(
            tmpdir,
            {"a.txt": "a=1", "b.txt": "b=1"},
            sources=["a.*", "b.*"],
            expected={"a": 1, "b": 1},
        )

    def test_it_substitutes_from_environment_variables(self, tmpdir):
        with setenv(FOO="bar"):
            self.check_loadoptions(
                tmpdir,
                {"a.txt": "a=1", "a.bar.txt": "a=2"},
                tags=["{FOO}"],
                expected={"a": 2},
            )

        with setenv(FOO="baz"):
            self.check_loadoptions(
                tmpdir,
                {"a.txt": "a=1", "a.bar.txt": "a=2"},
                tags=["{FOO}"],
                expected={"a": 1},
            )

    def test_it_allows_missing_environment_variables(self, tmpdir):
        assert "FOO" not in os.environ
        self.check_loadoptions(
            tmpdir,
            {"a.txt": "a=1", "a.bar.txt": "a=2"},
            tags=["{FOO}"],
            expected={"a": 1},
        )


class TestDictFromOptions:
    def test_it_splits_on_prefix(self):
        options = Options(FOO_BAR=1, FOO_BAZ=2, FOO_BAR_BAZ=3, BAR=4)
        assert dict_from_options("FOO_", options) == {"BAR": 1, "BAZ": 2, "BAR_BAZ": 3}

    def test_it_splits_recursively(self):
        options = Options(
            A_A=1,
            A_B_C_D=2,
            A_B_E=3,
            A_F_G_H=4,
            A_I=5,
            J_A=6,
        )
        assert dict_from_options("A_", options, recursive=True) == {
            "A": 1,
            "B": {"C": {"D": 2}, "E": 3},
            "F": {"G": {"H": 4}},
            "I": 5,
        }


@contextlib.contextmanager
def setenv(**kw):
    saved = {k: os.environ[k] for k in kw if k in os.environ}
    os.environ.update(kw)
    yield os.environ
    for k in kw:
        if k in saved:
            os.environ[k] = saved[k]
        else:
            del os.environ[k]
